@@ -58,6 +58,24 @@ def prepare_docker_image(image, name, tag):
 		image = client.images.pull(name, tag)
 	return client
 
+def _docker_run(client, volumes, environment, command):
+	container = None
+	try:
+		container = client.containers.create(
+			image = IMAGE, 
+			command = command,
+			volumes = volumes ,
+			environment = environment,
+			detach = True 
+		)
+		container.start()
+		logs = container.logs(stream = True)
+		for s in logs:
+			print(s.decode('utf-8').strip())	
+	finally:
+		container.remove() if container is not None else None
+
+
 @click.group()
 def cli():
 	f"""Kaen helps you train distributed PyTorch deep learning models in AWS, Azure, and Google Cloud.
@@ -67,13 +85,13 @@ def cli():
 	pass
 
 @cli.command("init")
-@click.option('--provider', required=True, type=click.Choice(['aws', 'azure', 'gcp', 'local'], case_sensitive=False))
+@click.option('--provider', required=False, default='local', show_default=True, type=click.Choice(['aws', 'azure', 'gcp', 'local'], case_sensitive=False))
 @click.option('--backend', default='docker', type=click.Choice(['docker', 'k8s'], case_sensitive=False), show_default=True)
 @click.option('--worker-instances', default='1', type=int, show_default=True)
 @click.option('--worker-instance-type', default='t3.micro', show_default=True)
 @click.option('--manager-instances', default='1', type=int, show_default=True)
 @click.option('--manager-instance-type', default='t3.micro', show_default=True)
-@click.option('--volume-size', default='8', type=int, show_default=True)
+@click.option('--volume-size', default='16', type=int, show_default=True)
 @click.option('--ssh-connection-timeout', default='5', type=int, show_default=True)
 @click.option('--ssh-connection-attempts', default='1', type=int, show_default=True)
 def init(provider, 
@@ -98,13 +116,13 @@ def init(provider,
 
 class DojoCLI(click.MultiCommand):
 	@cli.command("init")
-	@click.option('--provider', required=True, type=click.Choice(['aws', 'azure', 'gcp', 'local'], case_sensitive=False))
+	@click.option('--provider', required=False, default='local', show_default=True, type=click.Choice(['aws', 'azure', 'gcp', 'local'], case_sensitive=False))
 	@click.option('--backend', default='docker', type=click.Choice(['docker', 'k8s'], case_sensitive=False), show_default=True)
 	@click.option('--worker-instances', default='1', type=int, show_default=True)
 	@click.option('--worker-instance-type', default='t3.micro', show_default=True)
 	@click.option('--manager-instances', default='1', type=int, show_default=True)
 	@click.option('--manager-instance-type', default='t3.micro', show_default=True)
-	@click.option('--volume-size', default='8', type=int, show_default=True)
+	@click.option('--volume-size', default='16', type=int, show_default=True)
 	@click.option('--ssh-connection-timeout', default='5', type=int, show_default=True)
 	@click.option('--ssh-connection-attempts', default='1', type=int, show_default=True)
 	def init(provider, 
@@ -147,28 +165,33 @@ class DojoCLI(click.MultiCommand):
 						"KAEN_PROVIDER": provider,
 						}
 
-		volumes = {f"{os.getcwd()}" : {"bind": "/workspace", "mode": "rw"},}
+		# volumes = {f"{os.getcwd()}" : {"bind": "/workspace", "mode": "rw"},}
+
+		volumes = get_volumes(provider)
 		
-		if provider == 'local':
+		# if provider == 'local':
+		# 	
+		# 	volumes = {
+		# 		**volumes, 
+		# 		**{"/var/run/docker.sock": {"bind": "/var/run/docker.sock", "mode": "rw"}}
+		# 	}
+		# else:
+		if provider in ('local'):
 			print(f"Preparing a dojo using {backend} on {provider} ...")
-			volumes = {
-				**volumes, 
-				**{"/var/run/docker.sock": {"bind": "/var/run/docker.sock", "mode": "rw"}}
-			}
 		else:
-			environment = { **environment, **{
-							"KAEN_DOJO_CONNECT_TIMEOUT": ssh_connection_timeout,
-							"KAEN_DOJO_CONNECT_ATTEMPTS": ssh_connection_attempts,
-							"AWS_ACCESS_KEY_ID": os.environ['AWS_ACCESS_KEY_ID'],
-							"AWS_DEFAULT_REGION": os.environ['AWS_DEFAULT_REGION'],
-							"AWS_SECRET_ACCESS_KEY": os.environ['AWS_SECRET_ACCESS_KEY'],
-							"TF_VAR_instances": worker_instances,
-							"TF_VAR_worker_instance_type": worker_instance_type,
-							"TF_VAR_manager_instances": manager_instances,
-							"TF_VAR_manager_instance_type": manager_instance_type,
-							"TF_VAR_volume_size": volume_size,}
-			}		
 			print(f"Preparing a dojo using {backend} on {provider} with {worker_instances} worker(s) and {manager_instances} manager(s) ...")
+			environment = { **environment, **{
+								"KAEN_DOJO_CONNECT_TIMEOUT": ssh_connection_timeout,
+								"KAEN_DOJO_CONNECT_ATTEMPTS": ssh_connection_attempts,
+								"AWS_ACCESS_KEY_ID": os.environ['AWS_ACCESS_KEY_ID'],
+								"AWS_DEFAULT_REGION": os.environ['AWS_DEFAULT_REGION'],
+								"AWS_SECRET_ACCESS_KEY": os.environ['AWS_SECRET_ACCESS_KEY'],
+								"TF_VAR_instances": worker_instances,
+								"TF_VAR_worker_instance_type": worker_instance_type,
+								"TF_VAR_manager_instances": manager_instances,
+								"TF_VAR_manager_instance_type": manager_instance_type,
+								"TF_VAR_volume_size": volume_size,}
+			}
 
 		container = client.containers.run(	image = IMAGE, 
 			command=f"/bin/bash -c '/opt/kaen/{backend}/{provider}/dojo_init.sh'",
@@ -192,44 +215,89 @@ class DojoCLI(click.MultiCommand):
 			assert_aws_credentials()
 
 		client = prepare_docker()
-		volumes = get_volumes(provider)
+		# volumes = get_volumes(provider)
 
-		container = client.containers.run(	
-			image = IMAGE, 
-			command = f"/bin/bash -c '/opt/kaen/{backend}/{provider}/dojo_activate.sh'",
-			volumes = volumes ,
-			environment = { "KAEN_DOJO": dojo, 
-							},
-			auto_remove = True,
-			detach = True )
+		_docker_run(prepare_docker(), 
+								get_volumes(provider), 
+								{ "KAEN_DOJO": dojo,},
+								f"/bin/bash -c '/opt/kaen/{backend}/{provider}/dojo_activate.sh'")
 
-		logs = container.attach(stream = True, logs = True)
-		for s in logs:
-			print(s.decode('utf-8'))	
+		# container = client.containers.run(	
+		# 	image = IMAGE, 
+		# 	command = f"/bin/bash -c '/opt/kaen/{backend}/{provider}/dojo_activate.sh'",
+		# 	volumes = volumes ,
+		# 	environment = { "KAEN_DOJO": dojo, 
+		# 					},
+		# 	auto_remove = True,
+		# 	detach = True )
+
+		# logs = container.attach(stream = True, logs = True)
+		# for s in logs:
+		# 	print(s.decode('utf-8').strip())	
+
+		
 
 	@click.command()	
 	@click.option('--reverse', '-r',  is_flag = True, default = False)
 	def ls(reverse):		
 		"""Output a list with the training dojo(s) in the current workspace.		
-		By default, the dojo that was created first in the workspace is listed first."""
+		By default, the dojo that was created most recently in the workspace is listed first.
+		Use the --reverse option to reverse the order of the dojo listing."""
 
-		client = prepare_docker()
+		# client = prepare_docker()
 		
 		environment = {
-			"KAEN_DOJO_LS_REVERSE": "" if reverse else "-r"
+			"KAEN_DOJO_LS_REVERSE": "" if not reverse else "-r"
 		}
 
-		container = client.containers.run(	
-			image = IMAGE, 
-			command=f"/bin/bash -c '/opt/kaen/dojo_ls.sh'",
-			volumes = get_volumes() ,
-			environment = environment,
-			auto_remove = True,
-			detach = True )
+		_docker_run(prepare_docker(), 
+								get_volumes(), 
+								{"KAEN_DOJO_LS_REVERSE": "" if not reverse else "-r",},
+								f"/bin/bash -c '/opt/kaen/dojo_ls.sh'")
+
+		# container = None
+		# try:
+
+		# 	container = client.containers.create(
+		# 		image = IMAGE, 
+		# 		command = f"/bin/bash -c '/opt/kaen/dojo_ls.sh'",
+		# 		volumes = get_volumes() ,
+		# 		environment = environment,
+		# 		# auto_remove = True,
+		# 		detach = True 
+		# 	)
+		# 	container.start()
+		# 	logs = container.logs(stream = True)
+		# 	for s in logs:
+		# 		print(s.decode('utf-8').strip())	
+
+		# finally:
+		# 	container.remove() if container is not None else None
+
+		# container = client.containers.run(	
+		# 	image = IMAGE, 
+		# 	command = f"/bin/bash -c '/opt/kaen/dojo_ls.sh'",
+		# 	volumes = get_volumes() ,
+		# 	environment = environment,
+		# 	auto_remove = True,
+		# 	detach = True )
+		# logs = container.attach(stream = True, logs = True)
+		# for s in logs:
+		# 	print(s.decode('utf-8').strip())	
+
+		# result = client.containers.run(	
+		# 	image = IMAGE, 
+		# 	command = f"/bin/bash -c '/opt/kaen/dojo_ls.sh'",
+		# 	volumes = get_volumes() ,
+		# 	environment = environment,
+		# 	auto_remove = False,
+		# 	tty = True,
+		# 	stdout = True,
+		# 	stderr = True,
+		# 	stream = False,
+		# 	detach = True )			
+		# print(result.decode('utf-8').strip())
 
-		logs = container.attach(stream = True, logs = True)
-		for s in logs:
-			print(s.decode('utf-8').strip())	
 
 	@click.command()
 	def prune():
@@ -280,7 +348,7 @@ class DojoCLI(click.MultiCommand):
 
 		logs = container.attach(stream = True, logs = True)
 		for s in logs:
-			print(s.decode('utf-8'))	
+			print(s.decode('utf-8').strip())	
 	
 	@click.command()	
 	@click.argument('dojo')
@@ -291,21 +359,26 @@ class DojoCLI(click.MultiCommand):
 			assert_aws_credentials()
 
 		client = prepare_docker()
-		
-		container = client.containers.run(	
-			image = IMAGE, 
-			command=f"/bin/bash -c '/opt/kaen/dojo_inspect.sh'",
-			volumes = get_volumes(provider) ,
-			environment = { "KAEN_DOJO": dojo, 
-							},
-			auto_remove = True,
-			detach = True )
 
-		logs = container.attach(stream = True, logs = True)
-		for s in logs:
-			s = s.decode('utf-8').strip()
-			if len(s):
-				print(s)
+		volumes = get_volumes(provider)
+		environment = { "KAEN_DOJO": dojo, }
+		_docker_run(client, 
+									volumes, 
+									environment, 
+									f"/bin/bash -c '/opt/kaen/dojo_inspect.sh'")
+		
+		# container = client.containers.run(	
+		# 	image = IMAGE, 
+		# 	command=f"/bin/bash -c '/opt/kaen/dojo_inspect.sh'",
+		# 	volumes = ,
+		# 	environment = 
+		# 	auto_remove = False,
+		# 	detach = True )
+
+		# logs = container.attach(stream = True, logs = True, stdout = True, stderr = True)
+		# for s in logs:
+		# 	print(s.decode('utf-8').strip())
+		# container.remove()
 
 	def list_commands(self, ctx):
 		return ['ls', 'rm', 'prune', 'activate', 'inspect', 'init']
@@ -351,21 +424,34 @@ class JobCLI(click.MultiCommand):
 		client = prepare_docker()
 		print('done')
 
-		container = client.containers.run(image = IMAGE, 
-			command=f"/bin/bash -c '/opt/kaen/{backend}/{provider}/job_create.sh'",
-			volumes = get_volumes(provider) ,
-			environment = { "KAEN_DOJO": dojo,
+		volumes = get_volumes(provider)
+		environment = { "KAEN_DOJO": dojo,
 							"KAEN_JOB": job,
 							"KAEN_JOB_IMAGE": image,
 							"KAEN_JOB_SUBNET": subnet if subnet else None,
 							"KAEN_JOB_MANAGER_IP": manager_ip if subnet else None,
-							},
-			auto_remove = True,
-			detach=True)
+							}
+
+		_docker_run(client, 
+								volumes,
+								environment,
+								f"/bin/bash -c '/opt/kaen/{backend}/{provider}/job_create.sh'")
+
+		# container = client.containers.run(image = IMAGE, 
+		# 	command=f"/bin/bash -c '/opt/kaen/{backend}/{provider}/job_create.sh'",
+		# 	volumes = get_volumes(provider),
+		# 	environment = { "KAEN_DOJO": dojo,
+		# 					"KAEN_JOB": job,
+		# 					"KAEN_JOB_IMAGE": image,
+		# 					"KAEN_JOB_SUBNET": subnet if subnet else None,
+		# 					"KAEN_JOB_MANAGER_IP": manager_ip if subnet else None,
+		# 					},
+		# 	auto_remove = True,
+		# 	detach=True)
 		
-		logs = container.attach(stream = True, logs = True)
-		for s in logs:
-			print(s.decode('utf-8').strip())	
+		# logs = container.attach(stream = True, logs = True)
+		# for s in logs:
+		# 	print(s.decode('utf-8').strip())	
 		
 		print(job)			
 
@@ -377,19 +463,26 @@ class JobCLI(click.MultiCommand):
 		client = prepare_docker()
 
 		backend, provider = get_backend_provider(job = job)
-		
-		container = client.containers.run(	
-			image=IMAGE, 
-			command=f"/bin/bash -c '/opt/kaen/{backend}/{provider}/job_inspect.sh'",
-			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
-											"mode": "rw"}} ,
-			environment = { "KAEN_JOB": job, },
-			auto_remove = False,
-			detach=True )
 
-		logs = container.attach(stream = True, logs = True)
-		for s in logs:
-			print(s.decode('utf-8'))	
+		vol = get_volumes(provider)
+		env = { "KAEN_JOB": job, }
+		_docker_run(client,
+								vol,
+								env,
+								f"/bin/bash -c '/opt/kaen/{backend}/{provider}/job_inspect.sh'")
+		
+		# container = client.containers.run(	
+		# 	image=IMAGE, 
+		# 	command=f"/bin/bash -c '/opt/kaen/{backend}/{provider}/job_inspect.sh'",
+		# 	volumes = get_volumes(provider) ,
+		# 	environment = { "KAEN_JOB": job, },
+		# 	auto_remove = False,
+		# 	detach = True )
+
+		# logs = container.attach(stream = True, logs = True, stdout = True, stderr = True)
+		# for s in logs:
+		# 	print(s.decode('utf-8').strip())
+		# container.remove()
 
 	@click.command()
 	@click.argument('job')
@@ -415,17 +508,22 @@ class JobCLI(click.MultiCommand):
 
 		vol = get_volumes(provider)
 
-		container = client.containers.run(	image = IMAGE, 
-			command=f"/bin/bash -c '/opt/kaen/{backend}/{provider}/job_start.sh'",
-			volumes = vol ,
-			environment = env,
-			auto_remove = True,
-			detach = True)
+		_docker_run(client, 
+								vol,
+								env,
+								f"/bin/bash -c '/opt/kaen/{backend}/{provider}/job_start.sh'")
 
-		if not detach:
-			logs = container.attach(stream = True, logs = True)
-			for s in logs:
-				print(s.decode('utf-8').strip())	
+		# container = client.containers.run(	image = IMAGE, 
+		# 	command=f"/bin/bash -c '/opt/kaen/{backend}/{provider}/job_start.sh'",
+		# 	volumes = vol ,
+		# 	environment = env,
+		# 	auto_remove = True,
+		# 	detach = True)
+
+		# if not detach:
+		# 	logs = container.attach(stream = True, logs = True)
+		# 	for s in logs:
+		# 		print(s.decode('utf-8').strip())	
 
 	@click.command()	
 	@click.option('--reverse', '-r',  is_flag = True, default = False)
@@ -436,20 +534,27 @@ class JobCLI(click.MultiCommand):
 		client = prepare_docker()
 		
 		env = {
-			"KAEN_JOB_LS_REVERSE": "" if reverse else "-r"
+			"KAEN_JOB_LS_REVERSE": "" if not reverse else "-r"
 		}
-		
-		container = client.containers.run(	
-			image=IMAGE, 
-			command=f"/bin/bash -c '/opt/kaen/job_ls.sh'",
-			environment = env,
-			volumes = get_volumes(),
-			auto_remove = False,
-			detach = True )
 
-		logs = container.attach(stream = True, logs = True)
-		for s in logs:
-			print(s.decode('utf-8').strip())
+		volumes = get_volumes()
+
+		_docker_run(client, 
+								volumes,
+								env,
+								f"/bin/bash -c '/opt/kaen/job_ls.sh'")
+		
+		# container = client.containers.run(	
+		# 	image=IMAGE, 
+		# 	command=f"/bin/bash -c '/opt/kaen/job_ls.sh'",
+		# 	environment = env,
+		# 	volumes = get_volumes(),
+		# 	auto_remove = False,
+		# 	detach = True )
+
+		# logs = container.attach(stream = True, logs = True)
+		# for s in logs:
+		# 	print(s.decode('utf-8').strip())
 
 	@click.command()
 	@click.argument('job')
@@ -457,12 +562,14 @@ class JobCLI(click.MultiCommand):
 		"""Remove the specified job and associated resources."""
 		client = prepare_docker()
 		env = dict({ "KAEN_JOB": job,})
+		backend, provider = get_backend_provider(job = job)
+		
+		volumes = get_volumes(provider)
 
 		container = client.containers.run(	
 			image=IMAGE, 
 			command=f"/bin/bash -c '/opt/kaen/{backend}/{provider}/hpo_disable.sh'",
-			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
-											"mode": "rw"}} ,
+			volumes = volumes,
 			environment = env,
 			auto_remove = False,
 			detach = True )
@@ -474,8 +581,7 @@ class JobCLI(click.MultiCommand):
 		container = client.containers.run(	
 			image=IMAGE, 
 			command=f"/bin/bash -c '/opt/kaen/{backend}/{provider}/job_rm.sh'",
-			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
-											"mode": "rw"}} ,
+			volumes = volumes,
 			environment = env,
 			auto_remove = False,
 			detach = True )
@@ -564,9 +670,9 @@ class HpoCLI(click.MultiCommand):
 		backend, provider = get_backend_provider(job = job)
 
 		env = dict({ "KAEN_JOB": job,
+					 "KAEN_HPO_SEED": seed,
 					 "KAEN_HPO_IMAGE": image,
 					 "KAEN_HPO_JOB_RUNS": num_runs,
-					 "KAEN_HPO_SEED": seed,
 					 "KAEN_HPO_SERVICE_PREFIX": service_prefix,
 					 "KAEN_HPO_SERVICE_NAME": service_name,
 					 },
@@ -575,16 +681,21 @@ class HpoCLI(click.MultiCommand):
 
 		vol = get_volumes(provider)
 
-		container = client.containers.run(	image = IMAGE, 
-			command=f"/bin/bash -c '/opt/kaen/{backend}/{provider}/hpo_enable.sh'",
-			volumes = vol,
-			environment = env,
-			auto_remove = True,
-			detach = True)
+		_docker_run(client, 
+								vol, 
+								env,
+								f"/bin/bash -c '/opt/kaen/{backend}/{provider}/hpo_enable.sh'")
 
-		logs = container.attach(stream = True, logs = True)
-		for s in logs:
-			print(s.decode('utf-8').strip())	
+		# container = client.containers.run(	image = IMAGE, 
+		# 	command=f"/bin/bash -c '/opt/kaen/{backend}/{provider}/hpo_enable.sh'",
+		# 	volumes = vol,
+		# 	environment = env,
+		# 	auto_remove = True,
+		# 	detach = True)
+
+		# logs = container.attach(stream = True, logs = True)
+		# for s in logs:
+		# 	print(s.decode('utf-8').strip())	
 
 	@click.command()
 	@click.argument('job')
@@ -705,7 +816,7 @@ class IdeCLI(click.MultiCommand):
 		token = md5(str.encode(str(time.time()))).hexdigest()
 		port = IdeCLI.next_available_port()
 		
-		container = IdeCLI._run(image_name, image_tag, [('8888', port)], [('JPY_TKN', token)])
+		container = IdeCLI._run(image_name, image_tag, [('8888', port), ('5000', '5000'), ('5001', '5001')], [('JPY_TKN', token)])
 		logs = container.attach(stream = True, logs = True)
 		for s in logs:
 			s = s.decode('utf-8').strip()
